# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import numpy as np
import csv
import copy


def _checkcoordinates(coordinates):
    if not isinstance(coordinates, np.ndarray):
        raise TypeError(f'coordinates must be a numpy array')
    if not coordinates.ndim == 2:
        raise ValueError(f'coordinates must have 2 dimension (ndim=2)')
    if not coordinates.shape[1] == 3:
        raise ValueError(f'coordinates must be of shape '
                         f'(n_solutions_points, 3)')
    return (coordinates)


def _checknames(names):
    if not isinstance(names, list):
        raise TypeError(f'names must be a list of string')
    return(names)


def _checksubject(subject):
    if subject is not None:
        if not isinstance(subject, str):
            raise ValueError(f'Subject must be a string but type '
                             f'{type(subject)} was found.')
    return(subject)


def read_spi(filename, subject=None):
    """Create a SourceSpace instance from Cartool spi file.

    Parameters
    ----------
    filename : str or file-like
        The spi file to read.
    subject : str
        The subject used to create the source space.
    Returns
    -------
    SourceSpace : pycartool.source_space.SourceSpace
        The SourceSpace.
    """
    with open(filename) as f:
        reader = csv.reader(f, delimiter='\t')
        d = list(reader)
        names = [elem[-1] for elem in d]
        coord = [elem[:-1] for elem in d]
        coord = np.array(coord)
        coord = coord.astype(np.float)
        Source_Space = SourceSpace(names, coord,
                                   filename=filename,
                                   subject=subject)
    return(Source_Space)


def write_spi(filename, SourceSpace):
    """Write Cartool spi file.

    Parameters
    ----------
    filename : str or file-like
        The spi file to write.
    SourceSpace : pycartool.source_space.SourceSpace
        The SourceSpace to save.
    """
    names = SourceSpace.names
    x, y, z = SourceSpace.coordinates.T
    with open(filename, 'w', newline='') as f:
        writer = csv.writer(f, delimiter='\t')
        for s in enumerate(zip(x, y, z, names)):
            writer.writerow(s[1])


class SourceSpace(object):
    """Container for source space data.

    Parameters
    ----------
    names : :obj:`list` of :obj:`str`, length (n_sources)
        The solutions point names.
    coordinates : numpy.ndarray, shape (n_sources, 3)
        The solutions point names coordinates.
    subject : str
        Subject from who the source space was created.
    filename : str or file-like
        If loaded from a file, the corresponding filename.

    Attributes
    ----------
    n_sources : int
        Number of sources.
    names : :obj:`list` of :obj:`str`, length (n_sources)
        The solutions point names.
    coordinates : numpy.ndarray, shape (n_sources, 3)
        The solutions point names coordinates.
    subject : str
        Subject from who the source space was created.
    filename : str or file-like
        If loaded from a file, the corresponding filename.

    """
    def __init__(self, names, coordinates, subject=None, filename=None):
        _checkcoordinates(coordinates)
        _checknames(names)
        if not len(names) == coordinates.shape[0]:
            raise ValueError(f'coordinates and names dimensions must match'
                             f' but found {len(names)} names and '
                             f'{coordinates.shape[0]} solution points'
                             f' coordinates')
        else:
            self.n_sources = len(names)

        self.subject = subject
        self.names = names
        self.coordinates = coordinates
        self.filename = filename

    def __repr__(self):
        s = f'{self.n_sources} sources'
        if self.subject is not None:
            s += f', subject : {self.subject}'
        if self.filename is not None:
            s += f', filename : {self.filename}'
        return(f'<SourceSpace or {s}>')

    def get_coordinates(self):
        """Return a copy of sources coordinates."""
        return (copy.deepcopy(self.coordinates))

    def get_names(self):
        """Return a copy of sources names."""
        return(copy.deepcopy(self.names))

    def save(self, filename):
        """Write SourceSpace to Cartool spi file.

        Parameters
        ----------
        filename : str or file-like
            The spi file to write.

        """
        write_spi(filename, self)

    def get_center_of_mass(self, method='mean'):
        """Compute the center of mass of the Source Space.

        Parameters
        ----------
        method : str
            The method to use. Can be 'mean', or 'median'.

        Returns
        -------
        center_of_mass : numpy.ndarray, shape(3)
            The x,y,z coordinates of the center of mass.

        """
        if method == 'mean':
            center_of_mass = np.mean(self.coordinates, axis=0)
        elif method == 'median':
            center_of_mass = np.median(self.coordinates, axis=0)
        return(center_of_mass)



# -*- coding: utf-8 -*-
# Authors: Tanguy Vivier <tanguy.viv@gmail.com>
#          Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)

from mne.channels import make_dig_montage
import numpy as np

def read_xyz(filename):
    """Reads and convert xyz positions to a mne montage type

    Parameters
    ----------
    filename : str
        The filename of the xyz file.
    Returns
    -------
    montage : mne.channels.montage.Montage
        Montage for EEG electrode locations.
    """

    n = int(open(filename).readline().lstrip().split(' ')[0])
    coord = np.loadtxt(filename, skiprows=1, usecols=(0, 1, 2), max_rows=n)
    names = np.loadtxt(filename, skiprows=1, usecols=3, max_rows=n,
                       dtype=np.dtype(str))
    names = names.tolist()
    ch_pos = dict()
    for i, name in enumerate(names):
        ch_pos[name] = coord[i]
    montage = make_dig_montage(ch_pos=ch_pos, coord_frame='head')
    return montage



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import os
import pytest
import numpy as np

from ..io.inverse_solution import read_is
from ..source_estimate import read_ris, SourceEstimate
from ..regions_of_interest import RegionsOfInterest
from ..source_space import SourceSpace

dir_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(dir_path, 'data')


def generate_source_space(size):
    """Generate Source space."""
    source_names = [('S' + str(i)) for i in range(0, size)]
    source_coords = np.random.rand(size, 3)
    source_space = SourceSpace(source_names, source_coords)
    return(source_space)


def generate_source_estimate(n_sources, n_times, sfreq):
    """Generate Source Estimate."""
    sources_tc = np.random.rand(n_sources, 3, n_times)
    source_estimate = SourceEstimate(sources_tc, sfreq)
    return(source_estimate)


def generate_rois(n_roi, n_sources):
    """Generate Rois."""
    rois_names = [('R' + str(i)) for i in range(0, n_roi)]
    roi_indices = []
    for i in range(0, n_roi):
        indices = np.random.randint(0, n_sources, size=i+1).tolist()
        roi_indices.append(indices)
    regions_of_interest = RegionsOfInterest(rois_names, roi_indices)
    return(regions_of_interest)


@pytest.mark.skip(reason='Requires large file')
def test_read_is():
    """Test read_is."""
    file_path = os.path.join('test')
    inverse_solution = read_is(file_path)
    if not inverse_solution['is_type'] == 'IS03':
        raise AssertionError()


def test_read_ris():
    """Test read_ris."""
    file_path = os.path.join(data_path, 'sample_test_ris.ris')
    read_ris(file_path)


def test_write_ris():
    """Test write_is."""
    file_path = os.path.join(data_path, 'sample_test_write_ris.ris')
    sources_tc = np.random.rand(5000, 3, 2048).astype('float32')
    sfreq = 512
    subject = 'test_subject'
    source_estimate = SourceEstimate(sources_tc, sfreq, subject=subject)
    source_estimate.save(file_path)
    read_source_estimate = read_ris(file_path)
    if not (read_source_estimate.sources_tc == sources_tc).all():
        raise AssertionError()
    if not read_source_estimate.sfreq == sfreq:
        raise AssertionError()
    if not read_source_estimate.filename == file_path:
        raise AssertionError()


def test_compute_tc():
    """Test SourceEstimate.compute_tc."""
    source_space = generate_source_space(5000)
    source_estimate = generate_source_estimate(5000, 2048, 512)
    source_estimate.source_space = source_space
    tc_mean = source_estimate.compute_tc(method='mean')
    tc_med = source_estimate.compute_tc(method='median')
    tc_svd = source_estimate.compute_tc(method='svd')
    print(tc_mean.shape)
    if not tc_mean.shape == (3, 2048):
        raise AssertionError()
    if not tc_med.shape == (3, 2048):
        raise AssertionError()
    if not tc_svd.shape == (1, 2048):
        raise AssertionError()


def test_compute_rois_tc():
    """Test SourceEstimate.compute_rois_tc."""
    source_space = generate_source_space(5000)
    source_estimate = generate_source_estimate(5000, 2048, 512)
    source_estimate.source_space = source_space
    regions_of_interest = generate_rois(32, 5000)
    regions_of_interest.source_space = source_space
    rois_estimates = source_estimate.compute_rois_tc(regions_of_interest)
    if not rois_estimates.sources_tc.shape == (32, 1, 2048):
        raise AssertionError()
    if not rois_estimates.sfreq == 512:
        raise AssertionError()


def test_per_roi():
    """Test SourceEstimate.per_roi"""
    source_space = generate_source_space(5000)
    source_estimate = generate_source_estimate(5000, 2048, 512)
    source_estimate.source_space = source_space
    regions_of_interest = generate_rois(32, 5000)
    regions_of_interest.source_space = source_space
    rois_estimates = source_estimate.per_roi(regions_of_interest)
    if not len(rois_estimates) == 32:
        raise AssertionError()



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import os
import numpy as np
from ..source_space import SourceSpace, read_spi, write_spi

dir_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(dir_path, 'data')


def test_read_spi():
    """Test read_spi"""
    file_path = os.path.join(data_path, 'sample_test_spi.spi')
    Source_space = read_spi(file_path)
    if not len(Source_space.names) == 5006:
        raise AssertionError()
    if not Source_space.coordinates.shape == (5006, 3):
        raise AssertionError()


def test_SourceSpace():
    """Test SourceSpace instance and methods"""
    coordinates = np.random.rand(5000, 3)
    names = [('S' + str(i)) for i in range(0, 5000)]
    Source_space = SourceSpace(names, coordinates)

    new_coordinates = np.random.rand(5000, 3)
    Source_space.coordinates = new_coordinates
    if not (Source_space.coordinates == new_coordinates).all():
        raise AssertionError()

    new_names = [('F' + str(i)) for i in range(0, 5000)]
    Source_space.names = new_names
    if not (Source_space.names == new_names):
        raise AssertionError()

    subject = 'test_subject'
    Source_space.subject = subject
    if not (Source_space.subject == subject):
        raise AssertionError()


def test_SourceSpace_get_center_of_mass():
    coordinates = np.random.rand(5000, 3)
    mean = np.mean(coordinates, axis=0)
    med = np.median(coordinates, axis=0)
    names = [('S' + str(i)) for i in range(0, 5000)]
    Source_space = SourceSpace(names, coordinates)
    if not (Source_space.get_center_of_mass(method='mean') == mean).all:
        raise AssertionError()
    if not (Source_space.get_center_of_mass(method='median') == med).all:
        raise AssertionError()


def test_write_spi():
    """Test write_spi"""
    filename = os.path.join(data_path, 'sample_test_write_spi.spi')
    coordinates = np.random.rand(5000, 3)
    names = [('S' + str(i)) for i in range(0, 5000)]
    Source_space = SourceSpace(names, coordinates)
    write_spi(filename, Source_space)
    Read_Source_space = read_spi(filename)
    if names != Read_Source_space.names:
        raise AssertionError
    if (coordinates != Read_Source_space.coordinates).any():
        raise AssertionError



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import struct
import numpy as np


def read_is(filename):
    """Read Cartool inverse solution (.is) file.

    Parameters
    ----------
    filename : str or file-like
        the Inerse solution file (.is) to read.

    Returns
    -------
    inverse_solution : numpy.ndarray, shape (n_regularizations, n_dim, n_solutionpoints, n_channels)
        the inverse solution matrices. n_dim=1 if solutions are scalar or
        n_dim=3 for vectorial solutions.

    """
    with open(filename, 'rb') as f:
        print(f'Reading {filename}')
        print(f'Reading Header...')
        is_type = [struct.unpack('c', f.read(1))[0].decode('utf-8')
                   for i in range(4)]
        is_type = ''.join(is_type)
        if is_type not in ['IS01', 'IS02', 'IS03']:
            print(f'{is_type} : Invalid IS type, please check that '
                  'input file is a Inverse Solution matrix')
            raise ValueError
        print(f'IS type: {is_type}')
        n_channels = struct.unpack('I', f.read(4))[0]
        print(f'n_channels: {n_channels}')
        numsolutionpoints = struct.unpack('I', f.read(4))[0]
        print(f'n_solutionpoints: {numsolutionpoints}')
        numregularizations = struct.unpack('I', f.read(4))[0]
        print(f'n_regularizations: {numregularizations}')
        isinversescalar = struct.unpack('c', f.read(1))[0]
        if isinversescalar == b'\x01':
            n_dim = 1
            print(f'Inverse solution is Scalar')
        elif isinversescalar == b'\x00':
            print(f'Inverse solution is Vectorial')
            n_dim = 3
        else:
            raise ValueError(f'isinversescalar must be either 1 for scalar, '
                             f'either 0 for vectorial, but '
                             f'{ord(isinversescalar)} found.')

        if is_type in ['IS01', 'IS02']:
            buf = f.read(n_dim * numsolutionpoints * n_channel * 4)
            data = np.frombuffer(buf, dtype=np.float32)
            data = data.reshape(numsolutionpoints, ndim, n_channel)
            data = no.array([data])
            data = np.swapaxes(data, 1, 2)

        elif is_type == 'IS03':
            print(f"Reading Variable Header...")

            ch_names = []
            for _ in range(n_channels):
                name = [char for char in f.read(32).split(b'\x00')
                        if char != b''][0]
                ch_names.append(name.decode('utf-8'))

            solutionpoints_names = []
            for _ in range(numsolutionpoints):
                name = [char for char in f.read(16).split(b'\x00')
                        if char != b''][0]
                solutionpoints_names.append(name.decode('utf-8'))

            regularizations_values = []
            for _ in range(numregularizations):
                value = struct.unpack('d', f.read(8))[0]
                regularizations_values.append(value)
            print(f'Regularizations values: {regularizations_values}')

            regularizations_names = []
            for _ in range(numregularizations):
                name = [char for char in f.read(32).split(b'\x00')
                        if char != b''][0]
                regularizations_names.append(name.decode('utf-8'))
            print(f'Regularizations names: {regularizations_names}')

            regularisation_solutions = []
            buf = f.read(numregularizations
                         * n_dim
                         * numsolutionpoints
                         * n_channels
                         * 4)
            data = np.frombuffer(buf, dtype=np.float32)
            data = data.reshape(numregularizations, numsolutionpoints,
                                n_dim, n_channels)
            data = np.swapaxes(data, 1, 2)

    regularisation_solutions = np.array(regularisation_solutions)
    inverse_solution = {'is_type': is_type,
                        'is_scalar': True if isinversescalar == "0" else False,
                        'ch_names': ch_names,
                        'solutionpoints_names': solutionpoints_names,
                        'regularizations_values': regularizations_values,
                        'regularizations_names': regularizations_names,
                        'regularisation_solutions': data}
    return(inverse_solution)



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)

import os
from ..io.sef import read_sef, write_sef

dir_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(dir_path, "data")

datasef = os.path.join(data_path, "sample-sef.sef")


def test_read_sef():
    read_sef(datasef)
    return()


def test_write_sef():
    sef = read_sef(datasef)
    write_sef("test.sef", sef)
    return()



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import os

from mne.channels import DigMontage
from ..io.montage import read_xyz


dir_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(dir_path, "data")


def test_read_xyz():
    """Test read_xyz."""
    file_path = os.path.join(data_path, "EGI257.GenevaAverage13.10-10.xyz")
    montage = read_xyz(file_path)
    if not isinstance(montage, DigMontage):
        raise AssertionError()



# -*- coding: utf-8 -*-
# Authors: Tanguy Vivier <tanguy.viv@gmail.com>
#          Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import struct
import time
import datetime as dt
import numpy as np
import mne
from mne.io import RawArray
from mne import create_info


def read_sef(filename):
    """
    Reads file with format .sef, and returns a mne.io.Raw object containing
    the data.

    Parameters
    ----------
    filename : str or file-like
        The Simple EEG (.sef) file to read.

    Returns
    -------
    raw : mne.io.RawArray
        RawArray containing the EEG signals.
    """
    f = open(filename, 'rb')
    #   Read fixed part of the headerà
    version = f.read(4).decode('utf-8')
    if version != 'SE01':
        priint(f'Version : {version} not supported')
        raise ValueError()
    n_channels,         = struct.unpack('I', f.read(4))
    num_aux_electrodes, = struct.unpack('I', f.read(4))
    num_time_frames,    = struct.unpack('I', f.read(4))
    sfreq,              = struct.unpack('f', f.read(4))
    year,               = struct.unpack('H', f.read(2))
    month,              = struct.unpack('H', f.read(2))
    day,                = struct.unpack('H', f.read(2))
    hour,               = struct.unpack('H', f.read(2))
    minute,             = struct.unpack('H', f.read(2))
    second,             = struct.unpack('H', f.read(2))
    millisecond,        = struct.unpack('H', f.read(2))

    #   Read variable part of the header
    ch_names = []
    for _ in range(n_channels):
        name = [char for char in f.read(8).split(b'\x00')
                if char != b''][0]
        ch_names.append(name.decode('utf-8'))
    # Read data
    buffer = np.frombuffer(
        f.read(n_channels * num_time_frames * 8),
        dtype=np.float32,
        count=n_channels * num_time_frames)
    data = np.reshape(buffer, (num_time_frames, n_channels))
    # Create infos
    description = 'Imported with Pycartool'
    try:
        record_time = dt.datetime(year, month, day,
                                  hour, minute, second).timetuple()
        meas_date = (time.mktime(record_time), millisecond)
    except Exception as e:
        print('Cannot read recording date from file...')
        print(e)
        meas_date = None
    ch_types = ['eeg' for i in range(n_channels)]
    infos = create_info(ch_names=ch_names, sfreq=sfreq,
                        ch_types=ch_types)
    infos['description'] = description
    infos['meas_date'] = meas_date
    raw = RawArray(np.transpose(data), infos)
    return (raw)


def write_sef(filename, raw):
    """Export a raw mne file to a sef file.

    Parameters
    ----------
    filename : str or file-like
        Filename of the exported dataset.
    raw : mne.io.RawArray
        The raw data to export.
    """
    n_channels = len(raw.info['ch_names'])
    num_freq_frames = raw.n_times
    info = raw.info
    sfreq = info['sfreq']
    num_aux_electrodes = n_channels - len(mne.pick_types(info, meg=False,
                                                         eeg=True,
                                                         exclude=['']))
    f = open(filename, 'wb')
    f.write('SE01'.encode('utf-8'))
    f.write(struct.pack('I', n_channels))
    f.write(struct.pack('I', num_aux_electrodes))
    f.write(struct.pack('I', num_freq_frames))
    f.write(struct.pack('f', sfreq))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))
    f.write(struct.pack('H', 0))

    ch_names = info['ch_names']
    for k in range(n_channels):
        ch_name = ch_names[k]
        ch_name = ch_name.ljust(8)
        f.write(ch_name.encode('utf-8'))

    data = raw.get_data().astype(np.float32)
    data = np.reshape(data, n_channels * num_freq_frames, order='F')
    data.tofile(f)
    f.close()



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import struct
import numpy as np


def read_lf(filename):
    """Read Cartool leadfield matrix.

    Parameters
    ----------
    filename : str or file-like
        The leadfield file (.lf) to read.

    Returns
    -------
    leadfield_matrix : numpy.ndarray, shape (n_channels, n_sources, 3)
        the leadfield matrix.

    """
    with open(filename, 'rb') as f:
        byte = f.read(4)
        number_of_electrodes = struct.unpack('i', byte)[0]
        byte = f.read(4)
        number_of_solution_points = struct.unpack('i', byte)[0]
        buf = f.read(number_of_electrodes * number_of_solution_points*8)
        data = np.frombuffer(buf, dtype=np.double)
    number_of_points = int(number_of_solution_points/3)
    leadfield_matrix = data.reshape(number_of_electrodes, number_of_points, 3)
    return(leadfield_matrix)



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)

import os
from ..regions_of_interest import read_roi

dir_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(dir_path, "data")


def test_read_roi():
    """Test read_roi"""
    file_path = os.path.join(data_path, "sample_test_roi.rois")
    rois = read_roi(file_path)
    if not len(rois.names) == 32:
        raise AssertionError()
    if not len(rois.groups_of_indexes) == 32:
        raise AssertionError()



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
from .source_space import SourceSpace


def read_roi(filename, source_space=None):
    """Read Cartool region of interest (.rois) files.

    Parameters
    ----------
    filename : str or file-like
        The Regions Of Interest (.roi) file to read.

    source_space : pycartool.source_space.SourceSpace
        The SourceSpace corresponding to the regions of interest.

    Returns
    -------
    regions_of_interest : pycartool.regions_of_interest.RegionsOfInterest
        The regions of interest.

    Note
    -------
    Indices are broadcast to 0 based as expected in Python. In file, indices
    are saved starting from 1 as expected in Cartool.

    """
    with open(filename) as f:
        Roi_type = f.readline().strip()
        if Roi_type != 'RO01':
            print(f'{Roi_type} format not supported')
            raise ValueError
        print(Roi_type)
        n_orig = int(f.readline().strip())
        print(f'Dimension_of_original_data: {n_orig}')
        n_roi = int(f.readline().strip())
        print(f'Number of ROI: {n_roi}')
        rois_name = []
        rois_elements = []
        for _ in range(0, n_roi):
            roi_name = f.readline().strip()
            roi_elem = f.readline().split(' ')[:-1]
            # Use python 0base indices instead of Cartool 1 based indices
            roi_elem = [int(elem) - 1 for elem in roi_elem]
            rois_name.append(roi_name)
            rois_elements.append(roi_elem)
    if source_space is not None:
        if n_orig != source_space.n_sources:
            raise ValueError(f'Dimension of original data for roi file data is'
                             f' {n_orig}, but source space contains '
                             f'{source_space.n_sources} sources.')
    regions_of_interest = RegionsOfInterest(rois_name, rois_elements,
                                            source_space=source_space,
                                            filename=filename)
    return(regions_of_interest)


def _check_groups_of_indexes(groups_of_indexes):
    if not isinstance(groups_of_indexes, list):
        raise TypeError(f'groups_of_indexes must be a list of list')
    for e, elem in enumerate(groups_of_indexes):
        if not isinstance(elem, list):
            raise TypeError(f'groups_of_indexes must be a list of list but'
                            f' the {e} element is a {type(elem)}')


def _compute_number_of_sources(groups_of_indexes):
    groups_of_indexes_max = [max(elem) for elem in groups_of_indexes]
    maximum = max(groups_of_indexes_max)
    return(maximum)


class RegionsOfInterest(object):
    """Container for regions of interest.

    Parameters
    ----------
    names : :obj:`list` of :obj:`str`
        The regions of interest names.
    groups_of_indexes : :obj:`list` of :obj:`list` of :obj:`int`
        The sources indices belonging to each region of interest.
    source_space : pycartool.source_space.SourceSpace
        The source space associated to the regions of interest.
    filename : str or file-like
        The Regions Of Interest file (.roi) from which
        the data has been extracted.

    Attributes
    ----------
    names : :obj:`list` of :obj:`str`
        The regions of interest names.
    groups_of_indexes : :obj:`list` of :obj:`list` of :obj:`int`
        The sources indices belonging to each region of interest.
    source_space : pycartool.source_space.SourceSpace
        The source space associated to the regions of interest.
    filename : str or file-like
        The Regions Of Interest file (.roi) from which
        the data has been extracted.

    """
    def __init__(self, names, groups_of_indexes,
                 source_space=None, filename=None):
        _check_groups_of_indexes(groups_of_indexes)
        maximum = _compute_number_of_sources(groups_of_indexes)
        # Check that groups_of_indexes correspond to source space.
        if source_space is not None:
            if not isinstance(source_space, SourceSpace):
                raise TypeError(f'sourcespace must be an instance'
                                f' of SourceSpace.')
            if source_space.n_sources < maximum:
                raise ValueError(f'Indice {maximum} found in groups_of_indexes'
                                 f' but SourceSpace contains only'
                                 f' {source_space.n_sources} sources.')

        self.names = names
        self.groups_of_indexes = groups_of_indexes
        self.source_space = source_space
        self.filename = filename

    def __repr__(self):
        s = f'{len(self.names)} Rois'
        if self.filename is not None:
            s += f', filename : {self.filename}'
        return(f'<RegionsOfInterest or {s}>')



# -*- coding: utf-8 -*-
# Authors: Victor Férat <victor.ferat@live.fr>
#
# License: BSD (3-clause)
import struct
import numpy as np
from .source_space import write_spi, SourceSpace


def _check_method(method):
    if method not in ['svd', 'mean', 'median']:
        raise ValueError('Method must be either svd, mean, or median')


def _check_sources_tc(sources_tc):
    if not isinstance(sources_tc, np.ndarray):
        raise TypeError(f'sources_tc must be an instance of numpy.ndarray')
    if sources_tc.ndim != 3:
        raise ValueError(f'sources_tc must be of shape'
                         f'(n_solutionpoints, n_dim, n_timeframes)')
    if sources_tc.shape[1] not in [1, 3]:
        raise ValueError(f'sources_tc.shape[1] must be either 1 ( scalar) '
                         f'or 3 (vectorial)')
    return(sources_tc)


def read_ris(filename, source_space=None, subject=None):
    """Read Cartool Results of Inverse Solution computation (.ris) file.

    Parameters
    ----------
    filename : str or file-like
        the ris file to read.
    source_space : pycartool.source_space.SourceSpace
        The SourceSpace corresponding to the source estimate.
    subject : str or file-like
        The subject used to create the source estimate.
    Returns
    -------
    source_estimate : pycartool.source_estimate.SourceEstimate
        The SourceEstimate extracted from ris file.
    """
    with open(filename, 'rb') as f:
        print(f'Reading {filename}')
        print(f'Reading Header...')
        ris_type = [struct.unpack('c', f.read(1))[0].decode('utf-8')
                    for i in range(4)]
        ris_type = ''.join(ris_type)
        if ris_type not in ["RI01"]:
            raise ValueError(f'{ris_type} : Invalid RI type, please check that'
                             f' input file is a Result of Inverse Solution '
                             f' computation')
        print(f'IS type: {ris_type}')
        n_solutionpoints = struct.unpack('I', f.read(4))[0]
        print(f'n_solutionpoints: {n_solutionpoints}')
        n_timeframes = struct.unpack('I', f.read(4))[0]
        print(f'n_timeframes: {n_timeframes}')
        s_freq = struct.unpack('f', f.read(4))[0]
        print(f'Samplimg frequency: {s_freq}')
        isinversescalar = struct.unpack('c', f.read(1))[0]
        print(isinversescalar)
        if isinversescalar == b'\x01':
            n_dim = 1
            print(f'Result of Inverse Solution computation is Scalar')
        elif isinversescalar == b'\x00':
            print(f'Result of Inverse Solution computation is Vectorial')
            n_dim = 3
        else:
            raise ValueError(f'isinversescalar must be either 1 for scalar, '
                             f'either 0 for vectorial, but '
                             f'{ord(isinversescalar)} found.')
        buf = f.read(n_dim * n_solutionpoints * n_timeframes * 4)
        data = np.frombuffer(buf, dtype=np.float32)
        data = data.reshape(n_timeframes, n_solutionpoints, n_dim)
        data = np.swapaxes(data, 1, 2)
        source_estimate = SourceEstimate(data.T, s_freq,
                                         source_space=source_space,
                                         subject=subject,
                                         filename=filename)
        return(source_estimate)


def write_ris(source_estimate, filename):
    """Write SourceEstimate instance to file.

    Parameters
    ----------
    filename : str or file-like
        filename of the exported inverse solution computation.
    source_estimate : pycartool.source_estimate.SourceEstimate
        The SourceEstimate to save as a ris file.
    """
    data = source_estimate.sources_tc.T
    sfreq = source_estimate.sfreq
    if not isinstance(data, np.ndarray):
        raise TypeError('Input data must be a ndarray')
    if data.ndim != 3:
        raise ValueError('Input data must be a 3D array')
    if data.shape[1] == 1:
        isinversescalar = b'\x01'
    elif data.shape[1] == 3:
        isinversescalar = b'\x00'
    else:
        raise ValueError('Input data must have shape (_,1,_) (scalar)'
                         ' or (_,3,_) (vectorial)')
    ris_type = 'RI01'
    n_timeframes = data.shape[0]
    n_solutionpoints = data.shape[2]
    f = open(filename, 'wb')
    f.write(ris_type.encode('utf-8'))
    f.write(struct.pack('I', n_solutionpoints))
    f.write(struct.pack('I', n_timeframes))
    f.write(struct.pack('f', sfreq))
    f.write(struct.pack('c', isinversescalar))
    if isinversescalar == b'\x00':
        data = np.swapaxes(data, 1, 2)
    else:
        data = np.swapaxes(data, 0, 1)
    data = data.astype(np.float32)
    data.tofile(f)
    f.close()


class SourceEstimate(object):
    """Container for source estimate data.

    Parameters
    ----------
    sources_tc : numpy.ndarray, shape (n_solutionpoints, n_dim, n_timeframes)
        The sources time courses. Can be either scalar (ndim=1) or
        vectorial (ndim=3).
    sfreq : float
        The sampling frequency.
    source_space : pycartool.source_space.SourceSpace
        The SourceSpace corresponding to the source estimate.
    subject : str
        The subject used to create the source estimate.
    filename : str
        filename from which the source estimate was imported.

    Attributes
    ----------
    is_scalar : bool
        True if estimate is scalar, False is estimate is vectorial.
    n_sources : int
        Number of sources.
    sources_tc : numpy.ndarray, shape (n_solutionpoints, n_dim, n_timeframes)
        The sources time courses. Can be either scalar (ndim=1) or
        vectorial (ndim=3).
    source_space : pycartool.source_space.SourceSpace
        The SourceSpace corresponding to the source estimate.
    subject : str
        The subject used to create the source estimate.
    filename : str
        filename from which the source estimate was imported.

    """

    def __init__(self, sources_tc, sfreq,
                 source_space=None, subject=None, filename=None):
        _check_sources_tc(sources_tc)
        # Check that source estimate correspond to source space
        if source_space is not None:
            if not isinstance(source_space, SourceSpace):
                raise TypeError(f'sourcespace must be an instance'
                                f' of SourceSpace.')
            if not len(source_space.names) == sources_tc.shape[0]:
                raise ValueError(f'Expect {len(source_space.names)} time '
                                 f' courses from Source space, but found only '
                                 f'{sources_tc.shape[0]} in sources_tc')
        if sources_tc.shape[1] == 1:
            self.is_scalar = True
        elif sources_tc.shape[1] == 3:
            self.is_scalar = False

        self.n_sources = sources_tc.shape[0]
        self.sources_tc = sources_tc
        self.source_space = source_space
        self.sfreq = sfreq
        self.subject = subject
        self.filename = filename

    def __repr__(self):
        if self.is_scalar is True:
            s = f'Scalar, '
        else:
            s = f'Vectorial, '
        s += f'{self.n_sources} sources, sfreq : {self.sfreq}'

        if self.subject is not None:
            s += f', subject : {self.subject}'
        if self.filename is not None:
            s += f', filename : {self.filename}'
        return(f'<SourceEstimate or {s}>')

    def save(self, filename, export_spi=False):
        """Write SourceEstimate to Cartool ris file.

        Parameters
        ----------
        filename : str or file-like
            The ris file to write.
        export_spi : bool
            If True, also export the corresponding source space as ris file.
        """
        write_ris(self, filename)
        if export_spi is True:
            if self.source_space is not None:
                write_spi(filename[:-3] + '.spi', self.source_space)
            else:
                raise ValueError('Cannot save source space to file, source '
                                 'space is not defined')

    def per_roi(self, region_of_interest):
        """Short summary.

        Parameters
        ----------
        region_of_interest : pycartool.regions_of_interest.RegionsOfInterest
            The region of interest used to split the source estimate.

        Returns
        -------
        rois_source_estimate : :obj:`list` of :obj:`pycartool.source_estimate.SourceEstimate`
            A list of Source estimate instance restricted to a Region of
            interest

        """
        # Check is SourceSpace are the equals.
        if self.source_space is None:
            raise ValueError('Source spaces must be defined')
        if region_of_interest.source_space is None:
            raise ValueError('Source spaces must be defined')
        if self.source_space != region_of_interest.source_space:
            raise ValueError('Source estimate and Regions of interest must '
                             'share the same source space')

        rois_source_estimate = []
        rois_names = region_of_interest.names
        for r, _ in enumerate(rois_names):
            indices = region_of_interest.groups_of_indexes[r]
            roi_sources_tc = self.sources_tc[indices, :, :]
            roi_sources_pos = self.source_space.coordinates[indices]
            roi_sources_names = [self.source_space.names[i] for i in indices]
            source_space = SourceSpace(roi_sources_names, roi_sources_pos)
            source_estimate = SourceEstimate(roi_sources_tc, self.sfreq,
                                             source_space=source_space,
                                             subject=self.subject,
                                             filename=None)
            rois_source_estimate.append(source_estimate)
        return(rois_source_estimate)

    def compute_tc(self, method='svd'):
        """Short summary.

        Parameters
        ----------
        method : str
            the method use to compute the time course. Can be either 'mean',
            'median' or 'svd'. Default to 'svd'.

        Returns
        -------
        tc : numpy.ndarray, shape(n_dim, n_times)
            The global source estimate time course. Can be either Vectorial or
            Scalar depending of the source estimate and the method.

        """
        _check_method(method)
        if method == 'median':
            tc = np.median(self.sources_tc, axis=0)
        elif method == 'mean':
            tc = np.mean(self.sources_tc, axis=0)
        elif method == 'svd':
            n_times = self.sources_tc.shape[-1]
            sources_tc_flat = self.sources_tc.reshape(-1, n_times)
            U, s, V = np.linalg.svd(sources_tc_flat, full_matrices=False)
            scale = np.linalg.norm(s) / np.sqrt(len(sources_tc_flat))
            tc = np.array([scale * V[0]])
            pos = self.source_space.coordinates
            pos_flat = pos.reshape(-1)
            v = np.multiply(U[:, 0], pos_flat).reshape(-1, 3).mean(axis=0)
            v = v/np.linalg.norm(v)
        return(tc)

    def compute_rois_tc(self, region_of_interest, method='svd'):
        """Short summary.

        Parameters
        ----------
        region_of_interest : pycartool.regions_of_interest.RegionsOfInterest
            The region of interest used to split the source estimate.
        method : str
            the method use to compute the time course. Can be either 'mean',
            'median' or 'svd'. Default to 'svd'.

        Returns
        -------
        Roi_source_estimate : pycartool.source_estimate.SourceEstimate
            A source estimate instance where each source correspond to a region
            of interest.

        """
        rois_names = region_of_interest.names
        rois_estimates = self.per_roi(region_of_interest)
        rois_tc = np.array([roi.compute_tc() for roi in rois_estimates])
        rois_coordinates = np.array([estimate.source_space.get_center_of_mass()
                                     for estimate in rois_estimates])
        Roi_source_space = SourceSpace(rois_names, rois_coordinates)
        Roi_source_estimate = SourceEstimate(rois_tc, self.sfreq,
                                             source_space=Roi_source_space)
        return(Roi_source_estimate)



import setuptools
from os import path


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(

     name='pycartool',

     version='0.1.1c',

     author="Victor Férat",

     author_email="victor.ferat@live.fr",

     description="A simple open source Python package for I/O between Cartool"
                 " and MNE ( and more generally python)",

     long_description=long_description,

     long_description_content_type="text/markdown",

     url="https://github.com/Functional-Brain-Mapping-Laboratory/PyCartool",

     license="BSD-3-Clause",

     python_requires='>=3.6',

     install_requires=["mne", "numpy"],

     packages=setuptools.find_packages(exclude=['docs', 'tests']),

     classifiers=[

        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
 )
